from datetime import datetime

import yaml


def get_config(configs):
    """
        Gathers query configurations.yml file data.
    """
    with open(configs, 'r') as file:
        config = yaml.safe_load(file)
    return config


def date_to_unix(date_string, fmt='%Y-%m-%d'):
    """
        Takes a standard date string and converts it into unix timestamp.
    """
    date = datetime.strptime(date_string, fmt)
    return date.strftime("%s")


def config_constructor(app_id, config_path=None):
    """
        Constructs a valid json payload for the OneSignal method.
        :api_key: str. the api key is required in the payload
        :config_path: str. filepath to the config.yml file
    """
    payload = {'app_id': app_id}

    # only add configs if there is a path
    if config_path:
        configs = get_config(config_path)

        # ensure last_active_since is unix
        if 'last_active_since' in configs:
            configs['last_active_since'] = date_to_unix(configs['last_active_since'])

        payload.update(configs)
    return payload



import os
import pickle
import time

import googleapiclient
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from oauth2client.client import OAuth2WebServerFlow

from sdc_dp_helpers.api_utilities.retry_managers import request_handler
from sdc_dp_helpers.google_search_console.config_managers import get_config
from sdc_dp_helpers.google_search_console.utils import date_range, un_nest_keys
from sdc_helpers.slack_helper import SlackHelper


class CustomGSCReader:
    """
        Custom Google Search Console Reader
    """

    def __init__(self, config_path=None, **kwargs):
        if config_path is not None:
            self.config = get_config(config_path, fmt='yml')
            self.credentials = kwargs['credentials']

        if config_path is None:
            self.credentials = get_config(kwargs['credentials'], fmt='json')

        self.scopes = (
            'https://www.googleapis.com/auth/webmasters.readonly',
            'https://www.googleapis.com/auth/webmasters'
        )
        self.discovery_uri = (
            'https://www.googleapis.com/discovery/v1/apis/customsearch/v1/rest'
        )

        self.slack_webhook = kwargs.get('slack_webhook', None)

    def generate_authentication(self, auth_file_location='gsc_credentials.pickle'):
        """
        A user friendly method to generate a .pickle file for future authentication.
        For the first time, you would need to log in with your web browser based on
        this web authentication flow. After that, it will save your credentials in
        a pickle file. Every subsequent time you run the script, it will use the
        “pickled” credentials stored in credentials.pickle to build the
        connection to Search Console.
        """
        client_id = self.credentials['installed'].get('client_id')
        client_secret = self.credentials['installed'].get('client_secret')
        oauth_scope = 'https://www.googleapis.com/auth/webmasters.readonly'
        redirect_uri = 'urn:ietf:wg:oauth:2.0:oob'

        flow = OAuth2WebServerFlow(client_id, client_secret, oauth_scope, redirect_uri)
        authorize_url = flow.step1_get_authorize_url()
        print(f'Go to the following link in your browser: {authorize_url}')
        code = input('Enter verification code: ').strip()
        credentials = flow.step2_exchange(code)
        pickle.dump(credentials, open(auth_file_location, 'wb'))

    def webmasters_service(self):
        """
        Makes use of the .pickle cred file to establish a webmaster connection.
        """
        credentials = pickle.load(open(self.credentials, 'rb'))
        return build('webmasters', 'v3', credentials=credentials)

    @request_handler(
        wait=1, backoff_factor=5.0, backoff_method='random'
    )
    def run_request(self, service, request, site_url):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        print(f'Request: {request}')
        return service.searchanalytics().query(
            siteUrl=site_url,
            body=request
        ).execute()

    def query(self):
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from GSC API.
        """
        service = self.webmasters_service()
        start_date = self.config.get('startDate')
        end_date = self.config.get('endDate')

        data_set = []

        # split request by date to reduce 504 errors
        for date in date_range(start_date=start_date, end_date=end_date):

            # run until none is returned or there is no more data in rows
            row_index = 0
            while True:
                request = {
                    'startDate': date,
                    'endDate': date,
                    'dimensions': self.config.get('dimensions'),
                    'metrics': self.config.get('metrics'),
                    'searchType': self.config.get('searchType'),
                    'rowLimit': self.config.get('maxRows', 25000),
                    'startRow': row_index * self.config.get('maxRows', 25000)
                }

                try:
                    response = self.run_request(
                        service=service,
                        request=request,
                        site_url=self.config.get('siteUrl')
                    )

                    if response is None:
                        print('Response is None, stopping.')
                        break
                    if 'rows' not in response:
                        print('No more data in Response.')
                        break

                    # added additional data that the api does not provide
                    # un un_nest keys
                    for row in response['rows']:
                        row['site_url'] = self.config.get('siteUrl')
                        row['search_type'] = self.config.get('searchType')
                        row = un_nest_keys(
                            data=row,
                            col='keys',
                            key_list=self.config.get('dimensions'),
                            value_list=row.get('keys')
                        )
                        data_set.append(row)
                    row_index += 1

                except googleapiclient.errors.HttpError as error:
                    if error.resp.status == 403:
                        message = f"GSC quotaExceeded on: {self.config.get('siteUrl')} " \
                                  f"for {self.config.get('searchType')}, " \
                                  f"waiting for 15 minutes."
                        print(message)
                        if self.slack_webhook:
                            SlackHelper.send_message(
                                hook=self.slack_webhook,
                                color='#ffa500',
                                message=message
                            )
                        time.sleep(900)
                    else:
                        raise error

        return data_set



# pylint: disable=no-self-use

import datetime
import gzip
import json
import os
import time
import warnings

import requests
import yaml
from simplejson import JSONDecodeError

from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.onesignal.config_managers import config_constructor


class CustomOneSignalReader:
    """
        Custom OneSignal Reader
    """

    def __init__(self, api_key_path):
        self.api_key_path = api_key_path
        self.api_creds = self.get_api_key()

        self.app_id = self.api_creds.get('app_id')
        self.api_key = self.api_creds.get('api_key')

    def get_api_key(self) -> dict:
        """
            Gathers key and value pairs of key and their secret.
        """
        with open(self.api_key_path, 'r') as file:
            return yaml.safe_load(file)

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0.1)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 'random')
    )
    @retry_handler(
        exceptions=requests.exceptions.RetryError,
        total_tries=10,
        initial_wait=5.0,
        backoff_factor=2,
        should_raise=True
    )
    def get_csv_url_data(self, url):
        """
            Once the request is sent to retrieve the CSV url, a short period is needed
            before requesting the data from the server. It is handled here.
            :url: str.
        """
        # wait for csv to generate, this usually takes from a few seconds to half and hour
        while True:
            csv_response = requests.get(url)
            if csv_response.status_code == 403:
                raise requests.exceptions.RetryError('CSV File is still being generated or is not available')

            if csv_response.status_code == 200:
                # stop loop once csv is successfully generated
                break

        return gzip.decompress(csv_response.content)

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0.1)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 'random')
    )
    def get_csv_export(self, config_path=None, filter_now=True):
        """
            This method can be used to generate a compressed CSV
            export of all of your current user data.
            POST: https://onesignal.com/api/v1/players/csv_export?
            :filter_now: bool. Filter based on dynamic date, ie. yesterday.
        """

        url = 'https://onesignal.com/api/v1/players/csv_export'
        header = {'Authorization': f'Basic {self.api_key}'}
        json_payload = config_constructor(app_id=self.app_id, config_path=config_path)

        # if filter_now is true update last_active_since to current date
        if filter_now:
            yesterday = datetime.date.today() - datetime.timedelta(1)
            json_payload['last_active_since'] = yesterday.strftime("%s")

        print(f'Json Payload: {json_payload}')

        try:
            url_response = json.loads(
                requests.post(
                    url=url,
                    json=json_payload,
                    headers=header
                ).text
            )
            url = url_response.get('csv_file_url')

            return self.get_csv_url_data(url)

        except requests.exceptions.MissingSchema:
            # return nothing if there is no data for the given request
            return None

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0.1)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 'random')
    )
    def get_view_notifications(self, config_path=None):
        """
            View the details of multiple notifications.
            Returns offsets so this can be managed with the context.
            GET: https://onesignal.com/api/v1/notifications
            :config_path: str.
        """
        url = 'https://onesignal.com/api/v1/notifications'
        header = {
            'Content-Type': 'application/json',
            'Authorization': f'Basic {self.api_key}'
        }
        json_payload = config_constructor(app_id=self.app_id, config_path=config_path)

        print(f'Json Payload: {json_payload}')

        with requests.session() as session:
            response = session.get(
                url=url,
                json=json_payload,
                headers=header
            ).json()

            # loop through all available pages and apply them to a list
            data_set, total_count = list(), response.get('total_count')
            for offset in range(0, total_count):
                print(f'At offset: {offset}/{total_count}')
                json_payload['offset'] = offset

                response = session.get(
                    url=url,
                    json=json_payload,
                    headers=header
                )

                try:
                    response = response.json()
                    notifications = response.get('notifications', [])

                    if len(notifications) > 0:
                        # Only append if there is data in list
                        data_set.append(notifications[0])
                    else:
                        # Handle end of offset for extra security
                        # if there is no data, loop must break
                        break

                except JSONDecodeError as err:
                    warnings.warn(
                        f'Could not return response.\n{err}'
                    )

            return json.dumps(data_set)



# pylint: disable=too-few-public-methods
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import json
import os
from datetime import datetime

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, file_path, profile_name=None):
        self.bucket = bucket
        self.file_path = file_path

        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource('s3')

    def write_to_s3(self, json_data):
        """
            Writes json data to s3 and names it by
            date of writing. Also consumes expected metadata.json file.
        """
        timestamp_suffix = datetime.now().strftime('%Y_%m_%d_%H%M%S')
        key_path = f'{os.path.join(self.file_path, timestamp_suffix)}.json'

        self.s3_resource.Object(self.bucket, key_path) \
            .put(Body=json.dumps(json_data))



"""
    Utils functions
"""
import yaml


def init_config(file_name: str = './example.yaml'):
    """
        Util method that creates an example config at the given file_name.
        :file_name: Full path to config YAML file.
    """
    config = {
        "blast": {
            "status": "sent",
            "start_date": "yesterday",
            "end_date": "yesterday"
        },
        "list": {
            "start_date": "yesterday",
            "end_date": "yesterday"
        },
        "send": {
            "start_date": "yesterday",
            "end_date": "yesterday"
        }
    }

    # write to file
    with open(file_name, 'w') as file_:
        yaml.dump(config, file_)



"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import os
from typing import Callable

import httplib2
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from httplib2 import ServerNotFoundError
from oauth2client.service_account import ServiceAccountCredentials

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.api_utilities.date_managers import date_range, date_handler


class BaseGAV3Reader:
    """
    Version 3 Reporting API Reader class.
    Does not inherit GAV4 class in order to preserve all existing methods
    expected behaviours.
    """

    def __init__(self, config_file_path, creds_file_path, **kwargs):
        self.creds_file_path = creds_file_path
        self.config = load_file(config_file_path, 'yaml')

        self.service_account_email = kwargs.get(
            'service_account_email'
        )
        self.scopes = kwargs.get(
            'scopes', ['https://www.googleapis.com/auth/analytics.readonly']
        )

    @staticmethod
    def normalize_data(report):
        """
        Flattens the response data to a standard dict.
        """
        headers = [head['name'] for head in report['columnHeaders']]
        profile = report['profileInfo']

        data_set = []
        for row in report['rows']:
            tmp = dict(zip(headers, row))
            tmp.update(profile)
            data_set.append(tmp)

        return data_set

    def get_service(self):
        """
        Get a service that communicates to a Google API.
        """
        credentials = ServiceAccountCredentials.from_p12_keyfile(
            filename=self.creds_file_path,
            scopes=self.scopes,
            service_account_email=self.service_account_email
        )

        service = build(
            serviceName='analytics',
            version='v3',
            credentials=credentials
        )

        return service

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 0.01)
    )
    @retry_handler((TimeoutError, OSError, ServerNotFoundError, HttpError), total_tries=5)
    def run_request(self, service, start_date, end_date, view_id, start_index):
        """
        Separated query method to handle retry and delay methods.
        """
        print(f'Querying View ID: {view_id} at Dates: {start_date} to {end_date}.')
        print(f'At index: {start_index}')

        response = service.data().ga().get(
            ids=f'ga:{view_id}',
            start_date=start_date,
            end_date=end_date,
            # convert mets and dims into comma separated string
            metrics=','.join(self.config.get('metrics', [])),
            dimensions=','.join(self.config.get('dimensions', [])),
            sort=self.config.get('sort', None),
            filters=self.config.get('filters', None),
            max_results=self.config.get('max_results', 10000),
            start_index=start_index
        ).execute()

        return response

    def run_query(self):
        """
        Run Google Analytics query that queries per view_id and
        is able to batch by date.

        for report in ga_tap.run_query():
            print(report)

        :yield: dict. flat google analytics report dict
        """
        service = self.get_service()

        # batch dates defaults to true
        if self.config.get('batch_dates', True):
            # generate range of dates for date batch
            range_of_dates = date_range(
                start_date=self.config.get('start_date'),
                end_date=self.config.get('end_date')
            )

            # chunk dates
            for date in range_of_dates:
                for view_id in self.config.get('view_ids'):
                    page_token = 1  # handle paging of response
                    while True:
                        response = self.run_request(
                            service=service,
                            start_date=date,
                            end_date=date,
                            view_id=view_id,
                            start_index=page_token
                        )

                        if len(response.get('rows', [])) < 1:
                            print('No more data in response.')
                            break

                        page_token += response.get('itemsPerPage', 0)

                        yield self.normalize_data(response)

        else:
            start_date = date_handler(self.config.get('start_date'))
            end_date = date_handler(self.config.get('end_date'))

            for view_id in self.config.get('view_ids'):
                page_token = 1  # handle paging of response
                while True:
                    response = self.run_request(
                        service=service,
                        start_date=start_date,
                        end_date=end_date,
                        view_id=view_id,
                        start_index=page_token
                    )

                    if len(response.get('rows', [])) < 1:
                        print('No more data in response.')
                        break

                    page_token += response.get('itemsPerPage', 0)

                    yield self.normalize_data(response)


class BaseGAV4Reader:
    """
        Base Google Analytics V4 reader class.

        Used as basis for building custom GA V4 readers, it requires that you
        add your own Authentication through the auth_credentials method.
        The basic workflow is:
        (1) Create credentials
        (2) Build google service object
        (3) Return an iterable of the get_report method
        (4) Iterate until no pagination token is found. A.k.a. pageToken==None
    """
    analytics = None
    max_pages = None

    class ReaderIter:
        """Inner Class - Iterator that yields queries until max pages or pageToken == None."""
        page_counter = 0

        def __init__(
                self,
                *,
                query_fn: Callable,
                query: list,
                page_token: str = None,
                **kwargs
        ):
            self.query_fn = query_fn
            self.max_pages = kwargs.get("max_pages", None)
            self.page_token = page_token
            self.query = query

        def __iter__(self):
            """Iterator method for iterator class"""
            return self

        def __next__(self):
            """
                Next method for iterator class which includes
                the StopIteration criteria for iterable
            """
            # check StopIteration criteria
            if self.max_pages is not None:
                if self.page_counter >= self.max_pages:
                    print("stopping iteration on max pages = {}".format(self.max_pages))
                    raise StopIteration

            if self.page_counter > 0 and self.page_token is None:
                print("stopping iteration on pageToken = {}".format(self.page_token))
                raise StopIteration

            # run get_report
            report, self.page_token = self.query_fn(
                queries=self.query,
                page_token=self.page_token
            )
            print(
                'Finished query function for page {page} of view ID {view_id}'.format(
                    page=self.page_counter + 1,
                    view_id=self.query[0].get('viewId')
                )
            )

            # increment pagination counter
            self.page_counter += 1

            return report

    # pylint: disable=no-member
    def build_analytics_service_object(
            self,
            api_name: str = 'analyticsreporting',
            api_version: str = 'v4'
    ):
        """Initializes the analytics reporting service object.
        args:
        client_secrets_path (json) : Google API OAuth2 API credentials
        scopes (list or iterable) : scopes for Google OAuth2 authentication

        Returns:
        analytics an authorized analyticsreporting service object.
        """
        # authorize HTTP object
        http = self.credentials.authorize(http=httplib2.Http())
        # Build the service object.
        return build(api_name, api_version, http=http)

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 0.01)
    )
    def get_report(
            self,
            queries: list,
            page_token: str = None
    ):
        """
            Use the Analytics Service Object to query the Analytics Reporting API V4.

            args:
            analytics : analytics reporting service object
            query (list): array of queries. [max. 5]
            page_token (str): token sent denoting the last scraped data point
            return:
            analytics report data (json)

        """
        if page_token is not None:
            for query in queries:
                query.update({'page_token': page_token})

        # run batch query
        report = self.analytics.reports().batchGet(body={
            'reportRequests': queries}).execute().get('reports')
        page_token = report[0].get('nextPageToken')

        return report, page_token

    def run_query(self, query: list, page_token: str = None):
        """
            returns a sync iterator
        """
        iter_ = self.ReaderIter(
            query_fn=self.get_report,
            query=query,
            max_pages=self.max_pages,
            page_token=page_token)
        return iter_


class CustomGoogleAnalyticsReaderWithServiceAcc(BaseGAV4Reader):
    """
        Custom Google Analytics Reader which authenticates with Service account
    """

    # pylint: disable=super-init-not-called
    def __init__(
            self,
            service_account_secrets_path: str,
            service_account_email: str,
            scopes: list,
            **kwargs
    ):
        self.max_pages = kwargs.get("max_pages", None)
        self.credentials = self.auth_credentials(
            service_account_secrets_path, service_account_email, scopes)
        # build analytics object
        self.analytics = self.build_analytics_service_object()

    @staticmethod
    def auth_credentials(
            service_account_secrets_path,
            service_account_email,
            scopes
    ):
        """Get a service that communicates to a Google API.
        Args:
            service_account_secrets_path: The filepath to service secrets.
            api_name: The name of the api to connect to.
            api_version: The api version to connect to.
            scope: A list auth scopes to authorize for the application.
            key_file_location: The path to a valid service account p12 key file.
            service_account_email: The service account email address.
            scopes: Allowed scopes for the given credentials
        Returns:
            A service that is connected to the specified API.
        """
        return ServiceAccountCredentials.from_p12_keyfile(
            service_account_email, service_account_secrets_path, scopes=scopes)



import gzip
import json
import os
from datetime import timezone, datetime
from io import BytesIO
from pathlib import Path

import boto3
from interface import Interface


# pylint: disable=no-init, duplicate-code
class BaseWriter(Interface):
    """
        Base Writer Class with methods to implement when creating your own custom writer.
        Sets out a design pattern for creating the Custom writer class.

        Namely, the following methods should be implemented:
        (1) write metadata (optional)
        (2) append page
        (3) write file
    """

    def write_metadata(self, metadata: dict):
        """
            Method that writes metadata to local filesystem.
            args:
                metadata (dict) : metadata to be written
        """
        pass

    def append_page(self, report: list):
        """
            Called by main to get data and append rows to single array object.

            args:
                report (list): report data to append
        """
        pass

    def write_file(self):
        """Called to write a file or two database in bulk or s3"""
        pass


class CustomLocalJsonWriter:
    """
        Custom JSON Writer class which writes data to local filesystem.

        Writes data into a file.json per query. The files are written to a directory tree
    """

    def __init__(
            self,
            file_name: str,
            folder_path: str,
            **kwargs
    ):
        # creates full path
        self.folder_path = folder_path
        # create file paths
        self.file_name = file_name
        # currently overwrites metadata.json each time
        self.metadata_file = kwargs.get("metadata_file", "metadata.json")
        # init data
        self.data = []
        # init path
        self.full_path = None

    @staticmethod
    def mkdir_if_not_exists(full_path: str):
        """
            Make a new directory if one doesn't exist
        """
        Path(full_path).mkdir(parents=True, exist_ok=True)

    def check_files_in_dir(self):
        """
            Return the length of files in a directory
        """
        return len(os.listdir(self.full_path))

    @staticmethod
    def timestamp_suffix():
        """builds a filename suffix from the current timestamp"""
        curr_time = datetime.now().replace(tzinfo=timezone.utc).timestamp()
        suff = str(round(curr_time, 2)).replace(".", "_")
        return suff

    @staticmethod
    def flatten_multilist(mylist: list = None):
        """Utility function to flatten google analytics report list of list of dict"""
        flat_list = []
        for sublist in mylist:
            for item in sublist:
                flat_list.append(item)
        return flat_list

    def append_page(self, report: list):
        """
            Called by main to get data and append rows to single array object.
            args:
                report (dict): report data to append
        """
        # when to write (wrapper)
        data = [r.get('data', {}).get('rows', []) for r in report]
        restructured_data = self.flatten_multilist(data)
        self.data = self.data + restructured_data


class CustomS3JsonWriter(CustomLocalJsonWriter):
    """Class Extends Basic LocalJsonWriter"""

    def __init__(
            self,
            file_name: str,
            folder_path: str,
            bucket: str,
            profile_name: str = None,
            **kwargs
    ):
        self.os_path_sep = "/"

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.bucket = bucket
        """Writes a general object to s3"""
        self.s3_resource = self.boto3_session.resource('s3')

        super().__init__(file_name=file_name, folder_path=folder_path, **kwargs)

        # create directory structure if not exists
        self.full_path = self.set_full_path(folder_path)

        self.metadata_file = kwargs.get("metadata_file", "metadata.json.gz")

    @staticmethod
    def set_full_path(folder_path):
        """Set full path to write to"""
        # creates full path
        today_date = datetime.today()
        date_prefix = today_date.strftime('%Y/%m/%d')

        full_path = [date_prefix, folder_path]

        print(full_path)
        print(os.path.join(*full_path).replace('\\', '/'))
        return os.path.join(*full_path).replace('\\', '/')

    # pylint: disable=no-member,too-many-arguments
    def write_to_s3(
            self,
            json_data,
            data_path
    ):
        """
            Write report to s3 bucket
        """
        self.s3_resource.Bucket(self.bucket).put_object(
            Key=os.path.join(self.full_path, data_path).replace('\\', '/'),
            Body=json.dumps(json_data))

    def zip_write_to_s3(
            self,
            json_data,
            data_path):
        """
            Write compressed report to s3 bucket
        """

        gz_body = BytesIO()
        gz_file = gzip.GzipFile(None, 'wb', 7, gz_body)
        gz_file.write(json.dumps(json_data).encode('utf-8'))
        gz_file.close()

        self.s3_resource.Bucket(self.bucket).put_object(
            Key=os.path.join(self.full_path, data_path).replace('\\', '/'),
            ContentEncoding='gzip',
            Body=gz_body.getvalue()
        )

    def write_file(self, data: dict, compress=False):
        """
            Upload python dict into s3 bucket
            args:
                data (dict): report data to write
                compress (bool): compress data written to s3
        """

        # generate a unix timestamp suffix for file
        self.data = data
        if compress:
            self.zip_write_to_s3(self.data, self.file_name)
        else:
            self.write_to_s3(self.data, self.file_name)
        self.data = []



# pylint: disable=too-few-public-methods

"""
    CUSTOM WRITER CLASSES
"""
import os
from datetime import datetime

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, file_path, profile_name=None):
        self.bucket = bucket
        self.file_path = file_path

        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource('s3')

    def write_to_s3(self, data, extension):
        """
            Writes json data to s3 and names it by
            date of writing.
        """
        if extension not in ['json', 'csv', 'json.gz', 'csv.gz']:
            raise ValueError(f'Param: {extension} is not a '
                             f'valid option for extension.')

        now = datetime.now().strftime('%Y_%m_%d_%H%M%S')
        key_path = f'{os.path.join(self.file_path, now)}.{extension}'

        self.s3_resource.Object(self.bucket, key_path) \
            .put(Body=data)



import os
import warnings
from datetime import datetime
from pyspark.sql import functions as F
import pyspark.sql.types as T


def user_defined_timestamp(date_col, date_format: str = '%Y%m%d'):
    """
        (UDF) convert datetime column using python datetime

        args:
            date_col (pyspark.sql.functions.col): datetime column to format
            date_format (str): a datetime valid string format
        return
            (pyspark.sql.functions.col)
    """
    _date = datetime.strptime(date_col, '%Y-%m-%d')
    return _date.strftime(date_format)

def advanced_writer(
    *,
    dataframe,
    output_data_config: dict = None,
    **kwargs
):
    """
        Constructs a write function configuration and returns this.

        note:
            - on return write_function.csv() or write_function.json() or write_function.parquet()
    """
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements

    if not isinstance(output_data_config, dict):
        raise TypeError("output_data_config was not of type dict.")

    output_path =  output_data_config.get('output_path', None)
    if output_path is None:
        raise TypeError("output_path is required but was not found")

    output_type = output_data_config.get('output_type', None)
    if output_type is None:
        raise TypeError("output_type is required but was not found")

    partition_config = kwargs.get('partition_config', {})
    if not isinstance(partition_config, dict):
        raise TypeError("partition_config was not of type dict.")

    partitions =  partition_config.get('partitions', None)
    partition_by =  partition_config.get('partition_by', None)
    partition_type =  partition_config.get('partition_type', None)

    bucket_config = kwargs.get('bucket_config', {})
    if not isinstance(bucket_config, dict):
        raise TypeError("bucket_config was not of type dict.")

    bucket_type =  bucket_config.get('bucket_type', None)
    buckets =  bucket_config.get('buckets', None)

    compression_type = kwargs.get('compression_type', None)
    overwrite = kwargs.get('overwrite', False)
    header = kwargs.get('header', True)

    # output from this step is a dataframe writer
    if partition_type in ['key', 'random']:
        if partition_by is None:
            raise Exception("To perform repartitioning 'partition_by' is required.")

        col_structure = []
        for column in partition_by:
            col_structure.append(F.col(column))

        if partition_type == 'random':
            col_structure.append(F.rand)

        dataframe = dataframe.repartition(partitions, *col_structure)
        if bucket_type == 'size':
            # this will be ignored in the case where partitions are added
            dataframe_writer = dataframe.write.option(
                "maxRecordsPerFile", buckets
            ).partitionBy(*partition_by)
        else:
            dataframe_writer = dataframe.write.partitionBy(*partition_by)

    elif bucket_type == 'size':
        # this will be ignored in the case where partitions are added
        dataframe_writer = dataframe.write.option("maxRecordsPerFile", buckets)
    else:
        warnings.warn((
            'Defaulting to standard writer. '
            'Either set a num_partitions or max_size '
            'to bucket for additional options.'
        ))
        dataframe_writer = dataframe.write

    if overwrite is True:
        dataframe_writer = dataframe_writer.mode('overwrite')

    dataframe_writer = dataframe_writer.option("header", header)

    if compression_type is not None:
        dataframe_writer = dataframe_writer.option("compression", compression_type)

    if output_type == 'csv':
        dataframe_writer.csv(output_path)
    elif output_type == 'json':
        dataframe_writer.json(output_path)
    elif output_type == 'parquet':
        dataframe_writer.parquet(output_path)
    else:
        raise Exception(
            "Output type = {OUTPUT_TYPE} not supported. "
            "Only support csv and json at current.".format(
                OUTPUT_TYPE=output_type
            )
        )

def advanced_writer_w_custom_partition_naming(
    *,
    dataframe,
    output_data_config: dict,
    **kwargs
):
    """
        SDC custom pyspark writer that extends advanced_writer
        to offer special date partitioning which includes a more
        well formatted set of directory names for single level partition.
    """
    # pylint: disable=too-many-locals, too-many-function-args

    if not isinstance(output_data_config, dict):
        raise TypeError("output_data_config was not of type dict.")

    partition_config = kwargs.get('partition_config', {})
    if not isinstance(partition_config, dict):
        raise TypeError("partition_config was not of type dict.")

    partition_by =  partition_config.get('partition_by', None)
    partition_type =  partition_config.get('partition_type', None)
    date_partition_format = partition_config.get('date_partition_format', None)

    bucket_config = kwargs.get('bucket_config', {})
    if not isinstance(bucket_config, dict):
        raise TypeError("bucket_config was not of type dict.")

    if partition_type == 'date':
        output_path = output_data_config.get('output_path', None)
        if output_path is None:
            raise KeyError("output_path is required but was not found")

        output_type = output_data_config.get('output_type', None)
        if output_type is None:
            raise KeyError("output_type is required but was not found")

        if not isinstance(partition_by, str):
            raise TypeError(
                "For partition_type == date, "
                "the 'partition_by' field should be name of date field."
            )

        if date_partition_format is None:
            raise KeyError(
                "For partition_type == date, "
                "'date_partition_format' field is required but missing"
            )

        # pop partition_config from kwargs
        kwargs.pop('partition_config')

        # create udf to format date
        user_defined_timestamp_udf = F.udf(user_defined_timestamp, T.StringType())

        # create tmp_partition_field
        dataframe = dataframe.withColumn(
            'tmp_partition_field', user_defined_timestamp_udf(
                partition_by, F.lit(date_partition_format)
            )
        )

        partition_groups = list(
            dataframe.select(
                'tmp_partition_field'
            ).distinct().rdd.map(
                lambda r: r[0]
            ).collect()
        )

        for partition in partition_groups:
            partioned_dataframe = dataframe.filter(
                F.col("tmp_partition_field") == partition
            ).drop("tmp_partition_field")


            advanced_writer(
                dataframe=partioned_dataframe,
                output_data_config={
                   'output_path': os.path.join(output_path, partition),
                   'output_type': output_type
                },
                **kwargs
            )
    else:
        advanced_writer(
            dataframe,
            output_data_config=output_data_config,
            **kwargs
        )



from datetime import datetime, timedelta


def date_handler(date_string, date_format='%Y-%m-%d'):
    """
    Takes a date string or phrase and returns the valid date in
    the specified format.
    :date_string: A phrase like "3_days_ago", "yesterday" or "today"
                     or a date string such as "2021-01-01".
    :date_format: The returned formatting of the date.
    """
    try:
        if date_string == 'today':
            return datetime.now().strftime(date_format)
        if date_string == 'yesterday':
            return (datetime.now() - timedelta(days=1)).strftime(date_format)
        if '_days_ago' in date_string:
            return phrase_to_date(phrase=date_string, date_format=date_format)
        return date_string
    except ValueError as err:
        raise ValueError('StartDate requires a valid input '
                         'such as "today", "yesterday" or "<int>_days_ago".') from err


def phrase_to_date(phrase, date_format='%Y-%m-%d'):
    """
    Takes a typical phrase such as "3_days_ago" and returns a
    date based on that phrase.
    :phrase: str. Something like 3_days_ago or 25_days_ago etc.
    :return: '%Y-%m-%d'
    """
    try:
        date_delta = int(phrase.split('_')[0])
        return (datetime.now() - timedelta(days=date_delta)).strftime(date_format)
    except ValueError as err:
        raise ValueError(f'Phrasing for date: {phrase} is not valid, '
                         f'try something like: 3_days_ago') from err


def date_range(start_date, end_date, delta=timedelta(days=1)):
    """
    The range is inclusive, so both start_date and end_date will be returned.
    :start_date: The datetime object representing the first day in the range.
    :end_date: The datetime object representing the second day in the range.
    :delta: A datetime.timedelta instance, specifying the step interval. Defaults to one day.
    Yields:
        Each datetime object in the range.
    """

    start_date = date_handler(date_string=start_date)
    end_date = date_handler(date_string=end_date)

    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')

    current_date = start_date
    while current_date <= end_date:
        yield current_date.strftime('%Y-%m-%d')
        current_date += delta



import json

import yaml


def get_config(configs, fmt):
    """
        Gathers query configurations from json or yaml.
    """
    if configs is not None:
        if fmt in ['yaml', 'yml']:
            with open(configs, 'r') as file:
                config = yaml.safe_load(file)
        elif fmt == 'json':
            with open(configs, 'r') as file:
                config = json.load(file)

    return config



"""
    CUSTOM FALCON READER CLASSES
"""
import os
import time

import requests
import simplejson
import yaml

from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.api_utilities.date_managers import date_handler


class CustomFalconReader:
    """
        Custom Falcon Reader
    """

    def __init__(self, api_key_path, config_path=None):
        self.api_key_path = api_key_path
        self.config_path = config_path

        self.api_key = self.get_api_key()
        self.config = self.get_configuration()

        self.request_session = requests.Session()

    def get_api_key(self):
        """
            Gathers key and value pairs of client and their token.
        """
        with open(self.api_key_path, 'r') as file:
            data = yaml.safe_load(file)
            return data.get('api_key')

    def get_configuration(self):
        """
            Gathers optional and required configuration that
            defines the params for the api. This is optional.
            The configuration resembles the parameters of the
            given api method. eg:

            since: '2020-10-01T22:00:00.000Z'
            until: '2021-02-09T22:00:00.000Z'
            channels:
                - 'twitter_110592_91962761'
                - 'linkedin_110592_4840205'
            statuses:
                - 'all'
            limit: '30'
        """
        if self.config_path is not None:
            with open(self.config_path, 'r') as file:
                data = yaml.safe_load(file)
                return data
        return None

    def get_channel_ids(self):
        """
            Gather all available channel ids.
        """
        url = 'https://api.falcon.io/channels?apikey={apikey}' \
            .format(apikey=self.api_key)

        # no configs required, everything should be fetched, set limit to as high as possible
        try:
            response = requests.get(url=url, params={'limit': '999'}).json()
        except simplejson.errors.JSONDecodeError as err:
            raise EnvironmentError from err

        data_set = []
        for items in response['items']:
            data_set.append(items['id'])

        return data_set

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 0.01)
    )
    def _query_metrics_by_channel(self, url, params):
        """
        Separate initial request for the request handler.
        """
        return self.request_session.post(url=url, json=params)

    @retry_handler(exceptions=ConnectionRefusedError, total_tries=10)
    def metrics_by_channel(self, channel_ids, filter_type='all'):
        """
        Get metrics by channel Id context returns a request session with the ability
        to page with offsets.
        Content (or Post level) contains all metrics about your specific piece of
        content (posts). Here you will find impressions, reach, likes,
        shares and other metrics that show how well your specific post has performed.
        https://falconio.docs.apiary.io/reference/content-api/get-copied-content
        :channel_ids: list. List of channels for the API to return metrics on.
        :offset: int. Offset number from 0 until None is returned.
        :filter_type: str. filter type of 'all', 'default' or 'post'
        """
        url = 'https://api.falcon.io/measure/api/v1/content/metrics?apikey={apikey}' \
            .format(apikey=self.api_key)

        params = self.config
        params['channels'] = channel_ids

        # there are three options for dynamic time filtering
        if filter_type == 'default':
            params['since'] = date_handler(params['since'])
            params['until'] = date_handler(params['until'])
        elif filter_type == 'post':
            params['postsSince'] = date_handler(params['postsSince'])
            params['postsUntil'] = date_handler(params['postsUntil'])
        elif filter_type == 'all':
            params['since'] = date_handler(params['since'])
            params['until'] = date_handler(params['until'])
            params['postsSince'] = date_handler(params['postsSince'])
            params['postsUntil'] = date_handler(params['postsUntil'])
        else:
            raise ValueError('{val} is not a valid filter_today_type option'.format(
                val=filter_type
            ))

        offset, status_count, data_set = 0, 0, []
        while True:
            params['offset'] = offset
            print(f'Config: {self.config}')
            response = self._query_metrics_by_channel(url=url, params=params)

            # handle rate limit & rolling window limit with
            # a sleep method and counter before failure for non 429 responses
            if response.status_code == 500:
                # the api only returns 500 if a configuration is bad or invalid
                raise EnvironmentError('Status Code [500] returned.'
                                       'The reader configuration may be invalid.')

            if response.status_code == 429:
                print('Rolling Window Quota [429] reached, waiting for 15 minutes.')
                time.sleep(900)
            elif response.status_code != 200:
                raise ConnectionRefusedError

            data = response.json()
            print(f'Data: {data}')

            # break when there is no more data returned
            try:
                data_set.append(data[0])
                offset += 1
            except (IndexError, KeyError):
                print('No more data in response, stopping.')
                break

        return data_set
